from GOESLayer import GOESLayer
import sys
import os
from files.EMWINFile import EMWINFile
import logging
import re
from collections import deque
from datetime import datetime

##
# @brief EMWIN Layer class.
# @details Optional Sixth layer of LRIT packet processing. If an LRIT file
# turns out to actually be an EMWIN file, it needs to go through an additional
# layer of processing.
# @author Miguel Dominguez
# @date 2016
class EMWINLayer(GOESLayer):


   NULLS = re.compile('\x00{6}')
   FILENAME = re.compile('\/PF([\w]{8}.[\w]{3})')
   BLOCK_NUM = re.compile('\/PN[^\d]*([\d]+)')
   TOTAL_BLOCKS = re.compile('\/PT[^\d]*([\d]+)')
   CHECKSUM = re.compile('\/CS[^\d]*([\d]+)')
   TIMESTAMP = re.compile('\/FD([\d]+\/[\d]+\/[\d]+ [\d]+:[\d]+:[\d]+ [AP]M)')
   SPACE_PAD = re.compile('[ ]*\r\n')

   BLOCK_NUM_INDEX = 0
   TOTAL_BLOCKS_INDEX = 1
   CHECKSUM_INDEX = 2
   TIMESTAMP_INDEX = 3
   SPACE_PAD_INDEX = 4

   DATA_SUBGROUP = 1

   TIMESTAMP_FORMAT = '%m/%d/%Y %I:%M:%S %p'

   HEADER_SEQUENCE = [BLOCK_NUM,TOTAL_BLOCKS,CHECKSUM,TIMESTAMP,SPACE_PAD]

   FRAME_DATA_SIZE = 1024

   FILENAME_SIZE = 15

   HEADER_SIZE = 80

   NULL_BYTES_SIZE = 6
   ##
   # @fn __init__
   # @brief Constructor.
   # @details Initializes the logger and file table
   def __init__(self):
      GOESLayer.__init__(self)
      self.logger = logging.getLogger(__name__)
      self.fileTable = dict()
      self.dataBuffer = '' # data buffer is a string for regex purposes
      self.shippableProducts = deque()
      self.stateFunc = self.__processPreheader
      self.needsMoreData = False
      self.currentHeader = 0
      self.currentFilename = ''
      self.outputCount = 0
      self.stateBuffer = []

   #Searches entire data buffer
   def __processPreheader(self):
      match = self.NULLS.search(self.dataBuffer)
      if match is not None:
         self.logger.debug('Found Preheader null byte pattern')
         self.dataBuffer = self.dataBuffer[match.end():len(self.dataBuffer)]
         self.stateFunc = self.__processFilename
         self.needsMoreData = False
      else:
         self.needsMoreData = True

   def __processFilename(self):
      if self.dataBuffer >= self.FILENAME_SIZE:
         self.stateBuffer = self.dataBuffer[0:self.FILENAME_SIZE]
         self.logger.debug(self.stateBuffer)
         match = self.FILENAME.search(self.stateBuffer)
         #Found the filename adjacent to the null bytes as expected
         if match is not None:
            self.currentFilename = match.group(self.DATA_SUBGROUP)
            self.logger.debug('Found Filename {0}'.format(self.currentFilename))
            if self.currentFilename not in self.fileTable:
               self.fileTable[self.currentFilename] = EMWINFile()
               self.fileTable[self.currentFilename].filename = self.currentFilename
            self.stateFunc = self.__processHeader
            self.dataBuffer = self.dataBuffer[self.FILENAME_SIZE:len(self.dataBuffer)]
         #These must have been some other null bytes
         else:
            self.stateFunc = self.__processPreheader
         self.needsMoreData = False
      else:
         self.needsMoreData = True

   def __processHeader(self):
      remainingHeaderSize = self.HEADER_SIZE - self.FILENAME_SIZE
      if len(self.dataBuffer) >= remainingHeaderSize:
         self.stateBuffer = self.dataBuffer[0:remainingHeaderSize]
         match = self.HEADER_SEQUENCE[self.currentHeader].search(self.stateBuffer)
         #We're successfully finding the data we expect where we expect it
         if match is not None:
            if self.currentHeader == self.BLOCK_NUM_INDEX:
               currentBlock = int(match.group(self.DATA_SUBGROUP))
               self.fileTable[self.currentFilename].currentBlock = currentBlock
               self.currentHeader += 1
               self.logger.debug('Found Current Block {0}'.format(currentBlock))
            elif self.currentHeader == self.TOTAL_BLOCKS_INDEX:
               totalBlocks = int(match.group(self.DATA_SUBGROUP))
               self.fileTable[self.currentFilename].totalBlocks = totalBlocks
               self.currentHeader += 1
            elif self.currentHeader == self.CHECKSUM_INDEX:
               currentChecksum = int(match.group(self.DATA_SUBGROUP))
               self.fileTable[self.currentFilename].currentChecksum = currentChecksum
               self.currentHeader += 1
               self.logger.debug('Found Current Checksum {0}'.format(currentChecksum))
            elif self.currentHeader == self.TIMESTAMP_INDEX:
               timestamp = match.group(self.DATA_SUBGROUP)
               timestamp = datetime.strptime(timestamp,self.TIMESTAMP_FORMAT)
               self.fileTable[self.currentFilename].timestamp = timestamp
               self.currentHeader += 1
               self.logger.debug('Found Timestamp {0}'.format(timestamp))
            #Successfully full header
            elif self.currentHeader == self.SPACE_PAD_INDEX:
               self.currentHeader = 0
               self.stateFunc = self.__processData
               self.dataBuffer = self.dataBuffer[remainingHeaderSize:len(self.dataBuffer)]
               self.logger.debug('Found Space Padding')
         #EMWIN header was corrupt, go back to the beginning and try again
         else:
            self.stateFunc = self.__processPreheader
            self.needsMoreData = False
      else:
         self.logger.debug('NEEDS MORE DATA')
         self.needsMoreData = True

   #Assume the next 1024 bytes are data bytes
   def __processData(self):
      self.headerLen = 0
      if len(self.dataBuffer) >= self.FRAME_DATA_SIZE:
         self.logger.debug('Found Data')
         self.fileTable[self.currentFilename].currentData = [ord(x) & 0xFF for x in self.dataBuffer[0:self.FRAME_DATA_SIZE]]
         self.fileTable[self.currentFilename].verifyData()
         self.dataBuffer = self.dataBuffer[self.FRAME_DATA_SIZE:len(self.dataBuffer)]

         #Check if file is complete.  If so, output it
         if self.fileTable[self.currentFilename].isFileComplete():
            self.logger.debug('EMWIN File Completed!')
            self.shippableProducts.appendleft((self.fileTable[self.currentFilename],True))
            del self.fileTable[self.currentFilename]
         #If file is incomplete, check if it is too old.  If so, drop it
         elif self.fileTable[self.currentFilename].increaseAgeAndCheck():
            del self.fileTable[self.currentFilename]
         self.stateFunc = self.__processFooter
         self.needsMoreData = False
      else:
         self.needsMoreData = True


   def __processFooter(self):
      if len(self.dataBuffer) >= self.NULL_BYTES_SIZE:
         self.stateBuffer = self.dataBuffer[0:self.NULL_BYTES_SIZE]
         match = self.NULLS.search(self.stateBuffer)
         #I guess if we find it or don't it kind of doesn't matter at this point
         if match is not None:
            self.logger.debug('Found Footer')
            self.dataBuffer = self.dataBuffer[match.end():len(self.dataBuffer)]
         self.stateFunc = self.__processPreheader
      #Somehow the end got messed up
      else:
         self.needsMoreData = True

   ##
   # @fn threadFunction
   # @brief Custom thread function that allows multiple outputs per run.
   def threadFunction(self):
         while True:
            (inData,inOpsMsg) = self.inBuffer.get()
            if inData is not None:
               self.processData(inData,inOpsMsg)
               while self.shippableProducts:
                  self.outBuffer.put(self.shippableProducts.pop())
                  self.outputCount += 1
                  self.logger.info('EMWIN Output Count: {0}'.format(self.outputCount))
            else:
               self.outBuffer.put((None,None))
               break

   ##
   # @fn processData
   # @brief Extracts EMWIN files from LRIT files
   # @param[in] data LRIT File Object
   # @param[in] opsMsgs Information about CRC errors
   # @param[out] outData EMWIN File object
   # @param[out] outOpsMsg Filler, is ignored
   # @warning There are uncaught exceptions raised in processBuffer call here
   def processData(self,data,opsMsg):
     self.dataBuffer += ''.join([chr(x) for x in data.getRawData()])
     self.logger.debug(self.dataBuffer)
     self.needsMoreData = False
     while not self.needsMoreData:
         self.stateFunc()
         if self.shippableProducts:
            self.outputCount += 1
            self.logger.info('EMWIN Output Count: {0}'.format(self.outputCount))
            return self.shippableProducts.pop()
     return ([],False)
