#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 The Aerospace Corporation.
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
#

from __future__ import absolute_import

try:
    from enthought.traits.trait_base import ETSConfig
    ETSConfig.toolkit = 'qt4'
    # try Enthought Canopy package structure first
    from enthought.traits.api import *
    from enthought.traits.ui.api import *
    import enthought.traits.ui
    from enthought.pyface.action import *
    from enthought.pyface.api import *
except ImportError:
    from traits.trait_base import ETSConfig
    ETSConfig.toolkit = 'qt4'
    # try generic package structure
    from pyface.api import OK, FileDialog
    from traits.api import *
    from traitsui.api import *
    from traitsui.qt4 import enum_editor
    enum_editor.capitalize = lambda s: s
try:
    import ConfigParser as configparser #Py2.7 name
except:
    import configparser #Py3 name
import os
import os.path as osp
import sys
import threading as th
from threading import Timer
import json

## Used for some GUI changes and added features, as of 12/20/2018
import time
import weakref

import logging
logging.basicConfig(level=logging.WARN,
                    format='%(asctime)s - %(levelname)s - '
                           #'%(module)s.%(funcName)s -'
                           '%(message)s')
                    #datefmt='%m/%d/%Y %H:%M:%S.%f')    # comment out to use ISO8601 time format
logger = logging.getLogger(osp.splitext(osp.basename(__file__))[0])
logger.setLevel(logging.DEBUG)

#from .scan_read_files_tab import ScanReadFiles, ScanReadFilesTab
from .basic_configurations_tab import BasicConfigurationsTab
from .traitstools import ToggleButtonEditor, qt_to_traitsui_factory, QWidgetWrapper

from .goesradiomenu import top_menu_bar, TopMenuHandler
import config

KHZ, MHZ, GHZ = 1e3, 1e6, 1e9

from traitsui.key_bindings import KeyBinding, KeyBindings
keybindings = KeyBindings(
                KeyBinding(binding1='R', description='Reset error statistics',
                           method_name='reset_stats'),
                KeyBinding(binding1='T', description='Get error statistics',
                           method_name='trigger_stats'),
              )

from gnuradio.gr import pmt, top_block
class KBHandler(Handler):
    """Keybindings handler to reset error stats and trigger reports"""
    def reset_stats(self, info):
        obj = info.object
        if not obj.runfg: return
        obj.flowgraph.epy_block_1.to_basic_block()._post(pmt.intern('reset'), pmt.PMT_T)
        obj.flowgraph.ccsds_ccsds_decoder_0.to_basic_block()._post(pmt.intern('reset'), pmt.PMT_T)
        self.trigger_stats(info)    # reports reset back to the GUI

    def trigger_stats(self, info):
        obj = info.object
        if not obj.runfg: return
        obj.flowgraph.epy_block_1.to_basic_block()._post(pmt.intern('report'), pmt.PMT_T)
        obj.flowgraph.ccsds_ccsds_decoder_0.to_basic_block()._post(pmt.intern('report'), pmt.PMT_T)

from traitsui.tabular_adapter import TabularAdapter
class StatsAdapter(TabularAdapter):
    pass

class StatsDisplay(HasTraits):
    """Display for stat reports over ZMQ"""
    stat_tot_fer = Float(label='TOT FER')
    stat_rs_fer = Float(label='RS FER')
    stat_vit_ber = Float(label='VIT_BER')
    stat_miss_vcdu = Long(label='VCDU MISS')    # integers encoded as pmt long
    stat_tot_vcdu = Long(label='TOT VCDU')
    stat_lock_level = String("Unlocked", label='Lock Status')
    stat_snr = Float(-100, label='SNR (dB)', editor=TextEditor(format_str='%.1f'))
    stat_avgrate = Float(evaluate=lambda x: x / KHZ, label='Avg rate (Ksps)',
                    editor=TextEditor(format_func=lambda x:'%.1f' % (x / KHZ)))
    stat_instrate = Float(evaluate=lambda x: x / KHZ, label='Inst rate (Ksps)',
                    editor=TextEditor(format_func=lambda x:'%.1f' % (x / KHZ)))
    report_interval = Enum([0.5,] + range(1, 11), label='Report interval (s)')

    callback = Event
    reset = Event

    ErrItem = lambda name: Item(name, editor_args=dict(format_str='%6.2e'))
    traits_view = View(VGroup(HGroup('report_interval', '10',
                                     Label('Keypress: <b>t</b> = update error stats,   <b>r</b> = reset error stats')),
                              HGroup(ErrItem('stat_tot_fer'),
                                     ErrItem('stat_rs_fer'),
                                     ErrItem('stat_vit_ber'),
                                     'stat_miss_vcdu', 'stat_tot_vcdu', 'stat_lock_level',
                                     springy=True, style='readonly'),
                              HGroup('stat_snr', 'stat_avgrate', 'stat_instrate',
                                     springy=True, style='readonly')
                             ),
                        height=1.0
                      )

    def _report_interval_default(self): return 1

    def __init__(self):
        super(StatsDisplay, self).__init__()
        #self.on_trait_change(self.do_start, 'start', dispatch='new')
        #self.running = th.Event()

    def _reset_fired(self, value):
        """Reset statistics"""
        self.reset_traits([x for x in self.trait_names() if x.startswith('stat')])

    def _callback_fired(self, msg):
        """Process stats message from flowgraph"""
        # decode variously formatted pmts
        if pmt.is_real(msg):                # SNR message
            print("!!!! StatsAdapter, _callback_fired: conclude rx alive because reporting SNR")
            global deadman

            if (not ('deadman' in globals())):
                # Install deadman timer
                deadman = Deadman(5)

            # We're not locked up  
            deadman.resetTimer()

            self.stat_snr = pmt.to_float(msg)
        # goes error message (tuple)
        elif pmt.dict_has_key(msg, pmt.intern("goes errors")):
            errtup = pmt.to_python(pmt.dict_ref(msg, pmt.intern("goes errors"), pmt.PMT_NIL))
            self.trait_set(**dict(
                            zip(['stat_miss_vcdu', 'stat_tot_vcdu'], errtup)))
        elif pmt.is_pair(msg):              # rate message (dict)
            d = pmt.to_python(pmt.car(msg))
            if d == 'lock':
                is_locked = pmt.to_python(pmt.cdr(msg))
                self.stat_lock_level = "Locked" if is_locked else "Unlocked"
            elif d == 'viterbi_ber':
                self.stat_vit_ber = 10**pmt.to_python(pmt.cdr(msg))
            elif d == 'rs_frame_info':
                frame_info = pmt.to_python(pmt.cdr(msg))
                rs_rcvd_frames = frame_info[0]
                rs_frame_errors = frame_info[1]
                try:
                    self.stat_rs_fer = 1.0*rs_frame_errors / rs_rcvd_frames
                    self.stat_tot_fer = (rs_frame_errors + 4.0*self.stat_miss_vcdu) / ((self.stat_tot_vcdu+self.stat_miss_vcdu) * 4.0)
                except ZeroDivisionError:
                    self.stat_rs_fer = 0.0
                    self.stat_tot_fer = 0.0
            else:
                self.trait_set(**dict(stat_instrate=d['rate_now'], stat_avgrate=d['rate_avg']))
        else:
            pass    # TODO: log error?

# map osmo_sources to device arg strings
osmosourcemap = Trait('FILE',
                      {'FILE': "file=\'%s\',repeat=%s,rate=%d", # NOTE: need \' to allow files with spaces (see gr-osmocom arg_helpers.h)
                       #'RTL_TCP': 'rtl_tcp=%s',             # host:port, TODO: fix gr-osmocom rtl_tcp_source_f.cc
                       'RTL': 'rtl=%s',                     # 0-ref'd device index or serial
                       'UHD': 'uhd,addr=%s,otw_format=sc16'
                      }, section="Signal Source & Recording", config_name="Signal Source")

# import the flowgraph module
from ..flowgraphs import *  # import all the supported flowgraphs
from ..flowgraphs.utils import afoc, make_foffset_splitter
from PyQt4 import Qt


class StopReset(Exception):
    pass

class GNURadioInterface(HasTraits):

    container = None

    sourcefg = Instance(top_block, allow_none=True)
    flowgraph = Property(trait=Instance(QWidgetWrapper))
    stats_panel = Instance(StatsDisplay, (), transient=True)
    ## flowgraph parameters ##
    decim = Range(1, 10, 1, label='Decimation Factor',
                  section="Basic Parameters", config_name="Decimation Factor")
    samp_rate_MHZ = Property(Float, label='Sample rate (MHz)', depends_on='samp_rate')
    samp_rate = Float(2000000, section="Basic Parameters", config_name="Sample Rate")
    symb_rate = Float(927000, label='Symbol rate (Hz)',
                      section="Basic Parameters", config_name="Symbol Rate (Hz)")
    carrier_MHZ = Property(Float, label='Center Frequency (MHz)',  depends_on='carrier_fc')
    carrier_fc = Float(137500000, section="Basic Parameters", config_name="Center Frequency")
    digital_KHZ = Property(Float, label='Frequency Offset (KHz)', depends_on='digital_fc')
    digital_fc = Float(0, section="Basic Parameters", config_name="Frequency Offset (KHz)")
    enable_afoc = Bool(False, section="Basic Parameters", config_name="AFC Enabled?")
    xrit_imagery_filter = Bool(True, section="Signal Source & Recording", config_name="Save Xrit Imagery")
    emwin_filter = Bool(True, section="Signal Source & Recording", config_name="Save Emwin Products")
    dcs_filter = Bool(True, section="Signal Source & Recording", config_name="Save DCS Products")
    pulse_rolloff = Float(0.5, label='RRC Roll-Off', section="Basic Parameters", config_name="RRC Roll-off")
    gain = Range(0., 50.,section="Basic Parameters", config_name="Gain")
    ## flowgraph source parameters ##
    signal = Enum('LRIT', 'HRIT', value='LRIT', label='Signal',
                  section="Signal Source & Recording", config_name="Signal")
    # File trait returns unicode string which is incompatible with OSMOSDR source, use CStr
    recorded_data_file = CStr(label='File', editor=FileEditor(),
                              section="Signal Source & Recording", config_name="File Source")
    loop_file = Bool(True, label='Loop', section="Signal Source & Recording", config_name="Loop File?")
    device_addr = CStr('', section="Signal Source & Recording", config_name="Device Address")
    device_args = Property(Str)#, section="Signal Source & Recording", config_name="Device Arguments")
    osmo_source = osmosourcemap
    ## TODO: flowgraph output ##
    output_signals_to_file = Bool(False, label='Save signals',
                                  section="Signal Source & Recording", config_name="Output to file?")
    output_signals_to_file_dir = Directory(label='Dir',
                                           section="Signal Source & Recording", config_name="Output Directory")
    #lritdir_delegate = Instance(ScanReadFiles)
    #emwindir_delegate = Instance(ScanReadFiles)
    lritdir_delegate = CStr(label='Product Output Directory', editor=DirectoryEditor(invalid='directory_invalid'),
                     section="Signal Source & Recording", config_name="LRIT Output Directory")

    ## flowgraph run button and state
    runfg = Bool(False, editor=ToggleButtonEditor(unchecked_label='Start', checked_label='Stop'), transient=True)

    if os.environ['NO_GUI_FLOWGRAPH'] == '0':
        display_gui_flowgraph = Bool(True)
    elif os.environ['NO_GUI_FLOWGRAPH'] == '1':
		display_gui_flowgraph = Bool(False)
    else:
        raise ValueError('OS Environment Variable NO_GUI_FLOWGRAPH must be "0" or "1"')

    ## support for automatic frequency offset
    afoc_running = Any          # will be a threading.Event
    afoc_start = Event
    afoc_update = Event

    traits_view = View(
        VSplit(
            HGroup(
                VGroup(
                    Item('carrier_MHZ', editor_args=dict(auto_set=False,
                                                        format_str='%.3f')),
                    Item('samp_rate_MHZ', enabled_when='not runfg',
                                        editor_args=dict(auto_set=False)),
                    Item('symb_rate', editor_args=dict(auto_set=False)),
                    Item('decim', enabled_when='not runfg',
                                editor_args=dict(mode='spinner')),
                    HGroup(Item('digital_KHZ',
                                enabled_when='not enable_afoc',
                                editor_args=dict(auto_set=False,
                                                format_str='%.3f')),
                        Item('enable_afoc', label='AFC',
                                tooltip='Automatic Frequency Correction'),
                    ),
                    Item('pulse_rolloff', editor_args=dict(auto_set=False)),
                    Item('gain', style='simple', label='Gain (dB)'),
                    label='Basic parameters', show_border=True,
                ),
                VGroup(
                    Group(
                        Custom('osmo_source', label='Source'),
                        Custom('signal'),
                        HGroup(
                            Custom('loop_file', label='Loop file'),
                            UItem('recorded_data_file'),
                            enabled_when='osmo_source == "FILE"'
                        ),
                        Item('device_addr', label='Device addr',
                            editor_args=dict(auto_set=False),
                            #enabled_when='osmo_source != "FILE"'
                            ),
                        #HGroup(
                        #    Custom('output_signals_to_file'),
                        #    UItem('output_signals_to_file_dir', enabled_when='output_signals_to_file'),
                        #),
                        enabled_when='not runfg'
                    ),
                    Group(
                    Item('lritdir_delegate'),
                    Item('xrit_imagery_filter', label = 'XRIT Imagery Products'),
                    Item('emwin_filter', label = 'EMWIN Products'),
                    Item('dcs_filter', label = 'DCS Products'),
                    enabled_when='not runfg',
                    label='Product Output Directories',
                    show_border=True
                    ),
                    #UItem('device_args', label='Source string',
                    #      enabled_when='not runfg',
                    #      editor_args=dict(auto_set=False)),
                    UItem('runfg', resizable=True),
                    label='Signal source and recording', show_border=True,
                ),
                label='Settings', show_border=True
            ),
            UCustom('stats_panel'),
            Group(UCustom('flowgraph'), visible_when='display_gui_flowgraph')
        ),
        key_bindings=keybindings,
        handler=KBHandler()
    )

    def _recorded_data_file_default(self):
        return ''

    def __init__(self, container, *args, **kwds):
        super(GNURadioInterface, self).__init__(*args , **kwds)
        self.afoc_running = th.Event()
        self.container = weakref.ref(container)

        # trigger some updates in different threads
        self.on_trait_change(self.do_afoc, 'afoc_start', dispatch='new')
        self.on_trait_change(self.do_afoc_update, 'afoc_update', dispatch='ui')

    def __getstate__(self):
        """Overwrite the standard __getstate__ in order to filter out some\
           transient, but non-Traits member variables"""
        state=super(GNURadioInterface, self).__getstate__()
        for key in ['afoc_running', 'report_callback']:
            try:
                del state[key]
            except:
                pass
        return state

    def close(self, info, is_OK):
        self.runfg = False
        return True

    # couldn't add stats_panel.reset Event as handler for dynamic otc call, why?
    def _flowgraph_changed(self): self.stats_panel.reset = True

    def get_source(self):
        if self.osmo_source == 'FILE':
            return cfile_to_zmq_c(filepath=self.recorded_data_file,
                                  carrier_fc=self.carrier_fc,
                                  samp_rate=self.samp_rate,
                                  repeat=self.loop_file)
        if self.osmo_source == 'UHD':
            if self.device_addr:
                device_string = 'addr=%s' % self.device_addr
            else:
                device_string = ''
            return uhd_to_zmq_c(device_string=device_string,
                                carrier_fc=self.carrier_fc,
                                samp_rate=self.samp_rate,
                                gain=self.gain)
        if self.osmo_source == 'RTL':
            return rtl_to_zmq_c(carrier_fc=self.carrier_fc,
                                samp_rate=self.samp_rate,
                                gain=self.gain)

        raise NotImplementedError, 'Unsupported signal source: %s' % self.osmo_source

    # automagically cached until dependency changes.
    # NOTE: dependencies are those block parameters that are not updated when
    # the corresponding flowgraph attribute is updated, the parameter is instead
    # set at flowgraph construction.
    # dependency        block
    # -----------------------
    # decim           Frequency Xlating FIR filter
    # samp_rate       Frequency Xlating FIR filter
    @property_depends_on('signal, samp_rate, decim')
    def _get_flowgraph(self):
        if self.signal == 'LRIT':
            return QWidgetWrapper(qwidget=goeslritfg(
                       decim=self.decim,
                       carrier_fc=self.carrier_fc,
                       digital_fc=self.digital_fc,
                       samp_rate=self.samp_rate,
                       symb_rate=self.symb_rate,
                       # TODO: apparently get_flowgraph can be called b/4 delegate
                       # objects are constructed, why?
                       lritdir=self.lritdir_delegate or "",
                       emwindir=self.lritdir_delegate or "",
                       xrit_imagery_filter=self.xrit_imagery_filter,
                       emwin_filter=self.emwin_filter,
                       dcs_filter=self.dcs_filter,
                       report_callback=lambda m: self.stats_panel.trait_set(callback=m),
                       report_interval=self.stats_panel.report_interval))

        if self.signal == 'HRIT':
            return QWidgetWrapper(qwidget=goeshritfg(
                       decim=self.decim,
                       carrier_fc=self.carrier_fc,
                       digital_fc=self.digital_fc,
                       samp_rate=self.samp_rate,
                       symb_rate=self.symb_rate,
                       # TODO: apparently get_flowgraph can be called b/4 delegate
                       # objects are constructed, why?
                       lritdir=self.lritdir_delegate or "",
                       emwindir=self.lritdir_delegate or "",
                       xrit_imagery_filter=self.xrit_imagery_filter,
                       emwin_filter=self.emwin_filter,
                       dcs_filter=self.dcs_filter,
                       report_callback=lambda m: self.stats_panel.trait_set(callback=m),
                       report_interval=self.stats_panel.report_interval))


    @property_depends_on('osmo_source, recorded_data_file, '
                         'loop_file, samp_rate, device_addr', settable=True)
    def _get_device_args(self):
        if self.osmo_source == 'FILE':
            return self.osmo_source_ % (self.recorded_data_file,
                                        # w/o this mapping, boolean comes out
                                        # True/False which doesn't work with osmo source
                                        ('false', 'true')[self.loop_file],
                                        self.samp_rate)
        return self.osmo_source_ % (self.device_addr)

    # use default call to trigger update to device_args
    def _osmo_source_default(self): return 'FILE'

    # TODO: should all these be properties of the flowgraphs?  Probably
    # getter/setter for MHZ interface to carrier_fc
    # this was necessary to allow Hz entry for carrier_fc on CLI but MHZ in GUI
    def _get_carrier_MHZ(self): return self.carrier_fc / MHZ
    def _set_carrier_MHZ(self, new): self.carrier_fc = new * MHZ
    # same for digital_fc in KHZ
    def _get_digital_KHZ(self): return self.digital_fc / KHZ
    def _set_digital_KHZ(self, new): self.digital_fc = new * KHZ
    def _get_samp_rate_MHZ(self): return self.samp_rate / MHZ
    def _set_samp_rate_MHZ(self, new): self.samp_rate = new * MHZ

    def _runfg_changed(self, start):
        # TODO: shouldn't runfg be dependent on valid parameters?  Need validity
        # checks on lots of params and then an overall validity check to enable runfg button
        try:
            if start:
                self.sourcefg = self.get_source()
                #self.on_trait_change(self.sourcefg.set_carrier_fc, 'carrier_fc')
                #self.on_trait_change(self.sourcefg.set_gain, 'gain')
                self.flowgraph.start()
                self.sourcefg.start()
            else:
                # Cancel deadman switch
                global deadman
                if ('deadman' in globals()):
                    deadman.cancelTimer()
                    del deadman

                #self.on_trait_change(self.sourcefg.set_carrier_fc, 'carrier_fc', remove=True)
                #self.on_trait_change(self.sourcefg.set_gain, 'gain', remove=True)
                self.sourcefg.stop()#; print "SRC STOP"
                self.sourcefg.wait()#; print "SRC WAIT"
                self.flowgraph.stop()#; print "FG STOP"
                self.flowgraph.wait()#; print "FG WAIT"
                # TODO: this still freezes the GUI periodically, esp when hitting stop soon after start
                # NOTE: this appears to be fixed by adding finite LINGER to the
                # gr-zeromq base_impl constructor
                self.sourcefg = None#; print "SRC NONE"            # clean up here to avoid device lock preventing new sourcefg instance access
                #raise StopReset  # First attempt at clean reset after hitting stop button (1/10/2019)
        except (RuntimeError, NotImplementedError) as e:    # couldn't construct source
            logger.error('Error %s flowgraph: %s', 'starting' if start else 'stopping', e)
            self.runfg = False
        except AttributeError as e:                         # likely called stop w/o valid source
            logger.error('Error %s flowgraph: %s', 'starting' if start else 'stopping', e)

    # connect traits to flowgraph callbacks
    # NOTE: doing a dynamic on_trait_change setting creates a reference to
    # the sourcefg that hangs the UI when starting/stopping multiple times,
    # this static method avoids that issue; TODO: why???
    def _carrier_fc_changed(self, new):
        self.sourcefg and self.sourcefg.set_carrier_fc(new)
        self.flowgraph.set_carrier_fc(new)
        if self.container is not None:
            self.container().autosave_config()

    def _gain_changed(self, new):
        self.sourcefg and self.sourcefg.set_gain(new)
        if self.container is not None:
            self.container().autosave_config()

    def _digital_fc_changed(self, new):
        self.flowgraph.set_digital_fc(new)
        if self.container is not None:
            self.container().autosave_config()

    def _decim_changed(self, new):
        self.flowgraph.set_decim(new)
        if self.container is not None:
            self.container().autosave_config()

    def _samp_rate_changed(self, new):
        self.flowgraph.set_samp_rate(new)
        if self.container is not None:
            self.container().autosave_config()

    def _symb_rate_changed(self, new):
        self.flowgraph.set_symb_rate(new)
        if self.container is not None:
            self.container().autosave_config()

    @on_trait_change('lritdir_delegate')
    def set_lritdir(self, new):
        self.flowgraph.set_lritdir(new)
        if self.container is not None:
            self.container().autosave_config()

    #@on_trait_change('emwindir_delegate')
    #def set_emwindir(self, new): self.flowgraph.set_emwindir(new)

    @on_trait_change('stats_panel.report_interval')
    def set_report_interval(self, new):
        self.flowgraph.set_report_interval(new)
        if self.container is not None:
            self.container().autosave_config()

    @on_trait_change('xrit_imagery_filter')
    def set_xrit_imagery_filter(self, new):
        self.flowgraph.set_xrit_imagery_filter(new)
        if self.container is not None:
            self.container().autosave_config()

    @on_trait_change('emwin_filter')
    def set_emwin_filter(self, new):
        self.flowgraph.set_emwin_filter(new)
        if self.container is not None:
            self.container().autosave_config()

    @on_trait_change('dcs_filter')
    def set_dcs_filter(self, new):
        self.flowgraph.set_dcs_filter(new)
        if self.container is not None:
            self.container().autosave_config()

    def _lritdir_delegate_default(self):
        if osp.isdir(config.LRIT_DIRECTORY):
            return config.LRIT_DIRECTORY
        else:
            return config.HOME_DIRECTORY

    # this allows starting/stopping AFOC before or after flowgraph started/stopped
    @on_trait_change('runfg, enable_afoc')
    def trigger_afoc(self):
        if self.runfg and self.enable_afoc and not self.afoc_running.is_set():
            self.afoc_start = True
            self.afoc_running.set()
        else:
            self.afoc_running.clear()

    # start the automatic frequency offset compensation
    def do_afoc(self, go):
        if not go: return

        global logger
        # TODO: better way than to reach into flowgraph?
        fcn_meas_foff = self.flowgraph.probe_foffset.level
        fcn_set_foff = lambda fo: self.trait_set(afoc_update=fo)
        afoc(fcn_meas_foff, fcn_set_foff, self.afoc_running, logger=logger)

    # TODO: does this have to be dispatched in the ui thread?
    def do_afoc_update(self, new):
        dhw, ddg = make_foffset_splitter(self.samp_rate)(self.digital_fc, new)
        #print 'AFOC update', new, dhw, ddg
        self.carrier_fc += dhw
        self.digital_fc += ddg

# Deadman switch to restart flow graph if data production from signal source stops.
# Loss of data currently is defined when SNR estimates are no longer produced
# for greater than specified seconds.
class Deadman:
    def countdown(self):
        self.timer = Timer(self.trig_interval_sec, self.callback)
        self.timer.start()

    def __init__(self, trig_interval_sec):
        # print("!!!! Deadman, __init__: trig_interval_sec %f" % trig_interval_sec)
        self.trig_interval_sec = trig_interval_sec
        self.countdown()

    def callback(self):
        global grIntf
        print("!!!! Deadman: timeout triggered:", grIntf.runfg)

        # Stop receiver
        os._exit(1)

    def cancelTimer(self):
        self.timer.cancel()
        del self.timer
        #print("!!!! Deadman: timer cancelled")

    def resetTimer(self):
        self.cancelTimer()
        self.countdown()
        #print("!!!! Deadman: timer reset")

class Container(HasTraits):

    ## Attempt at autosave feature, as of 12/20/2018
    if os.environ['AUTOSAVE'] == "1":
        autosave = True
    else:
        autosave = False

    #scan_read_files_tab = Instance(ScanReadFilesTab, ())
    advanced_settings = Instance(GNURadioInterface, ())
    config_filename = File(osp.join(config.CONFIG_DIRECTORY, "default.conf"), transient=True)

    def autosave_config(self):
        if self.autosave:
            self.save_config(config_filename=self.config_filename)

    def save_config(self, config_filename=None):
        """TODO"""
        if config_filename is None:
            dialog = FileDialog(action='save as', default_path=self.config_filename)
            if dialog.open() == OK:
                config_filename = dialog.path
            else:
                config_filename = ""
        if osp.isdir(osp.dirname(config_filename)):
            self.config_filename=config_filename
        else:
            return
            #raise IOError("Invalid configuration file: %s"%config_filename)

        def get_full_state(trait_item, conf_par):
            """TODO"""
            state = trait_item.__getstate__()
            for key in state:
                item = state[key]
                # If the item is a HasTraits object, recurse into it
                if isinstance(item, HasTraits):
                    get_full_state(item, conf_par)
                else:
                    # Otherwise, try to get the configuration section and name
                    try:
                        section = trait_item.trait(key).section
                        config_name = trait_item.trait(key).config_name
                        if section is None:
                            raise AttributeError("%s has no configuration information"%key)
                        if isinstance(item, list):
                            item = json.dumps(item)
                        try:
                            # Now add the parameter to the configuration
                            conf_par.set(section, config_name, item)
                        except configparser.NoSectionError:
                            # Or add the section first if we need to
                            conf_par.add_section( section )
                            conf_par.set(section, config_name, item)
                    except AttributeError:
                        #raise AttributeError(("Item %s has no key: "%(item, key))
                        pass

        config = configparser.RawConfigParser()
        config.optionxform=str
        get_full_state(self, config)
        with open(self.config_filename, 'wb') as configFile:
            config.write(configFile)

    def load_config(self, config_filename=None):
        """TODO"""
        if config_filename is None:
            dialog = FileDialog(action='open', default_path=self.config_filename)

            if dialog.open() == OK:
                config_filename = dialog.path
                #print("Setting new config filename path", config_filename)
            else:
                config_filename = ""
        if not osp.isfile(config_filename): return
        self.config_filename=config_filename

        def set_full_state(trait_item, conf_par):
            """TODO"""
            state = trait_item.__getstate__()
            new_state = {}
            for key in state:
                item = state[key]
                if isinstance(item, HasTraits):
                    set_full_state(item, conf_par)
                else:
                    try:
                        section = trait_item.trait(key).section
                        config_name = trait_item.trait(key).config_name
                        if section is None:
                            raise AttributeError("%s has no configuration info (%s : %s)"%(key, section, config_name))
                        try:
                            if isinstance(item, bool):
                                new_item = conf_par.getboolean(section, config_name)
                            elif isinstance(item, float):
                                new_item = conf_par.getfloat(section, config_name)
                            elif isinstance(item, int):
                                new_item = conf_par.getint(section, config_name)
                            elif isinstance(item, list):
                                new_item = conf_par.get(section, config_name)
                                new_item = map(str, json.loads(new_item))
                            else:
                                new_item = conf_par.get(section, config_name)
                            new_state[key] = new_item
                        except configparser.NoSectionError:
                            print("Configuration not defined!", config_name)
                    except AttributeError:
                        pass
            trait_item.trait_set(**new_state)

        config = configparser.RawConfigParser()
        config.optionxform=str
        config.read(self.config_filename)
        set_full_state(self, config)

    view = View(
        Tabbed(
            Group(Item('advanced_settings', style='custom', show_label=False),
                  label = 'Receiver Control'),
            #Group(Item('scan_read_files_tab', style='custom', show_label=False),
                 # label = 'Scan and Read Files'),
        ),
        menubar=top_menu_bar,
        handler=TopMenuHandler(),
        resizable=True,
        buttons=[],
        title='xRIT / EMWIN Receiver',
    )

    # init accepts kwds for the GNURadioInterface (i.e. the passed command line args)
    def __init__(self, gr_kwds=None):
        HasTraits.__init__(self)
        self.advanced_settings = GNURadioInterface(
            self,
            #lritdir_delegate=self.scan_read_files_tab.first,
            #emwindir_delegate=self.scan_read_files_tab.second,
            **gr_kwds)

        if os.environ['AUTOSAVE'] == "1" and os.environ['CONFIG_FILE'] == "":
            os.environ['CONFIG_FILE'] = "autosave.conf"

        if os.environ['CONFIG_FILE'] != "":
            config_filename = osp.join(config.CONFIG_DIRECTORY, os.environ['CONFIG_FILE'])
            if osp.isfile(config_filename):
                self.config_filename = config_filename
                self.load_config(config_filename=self.config_filename)
            else:
                self.load_config(config_filename=self.config_filename)
                self.config_filename = config_filename
                self.save_config(config_filename=self.config_filename)
        else:
            self.load_config(config_filename=self.config_filename)

        if os.environ['AUTOSTART'] == '1':
            self.advanced_settings.runfg = True
        global grIntf
        grIntf = self.advanced_settings


def main(*args, **kwds):
    container = Container(gr_kwds=kwds)
    container.configure_traits()
    #Container(gr_kwds=kwds).configure_traits()

if __name__ == "__main__":
    sys.exit(main(*sys.argv[1:]))

