#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2016 The Aerospace Corporation.
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
#

import __builtin__ as btin
import os
import os.path as osp

import struct
import warnings
from collections import namedtuple

import lritheader as lrh

################################################################################
#                                                                              #
# Files section                                                                #
#                                                                              #
################################################################################

# File types
(RESERVED_FTYPE,
 IMAGE_DATA_FTYPE,
 MESSAGES_FTYPE,
 TEXT_FTYPE,
 ENC_KEY_FTYPE,
 METEORLOGICAL_FTYPE) = range(-1, 4) + [128,]

# TODO: use Table 16 as a verification for header compatibility?

################################################################################
#                                                                              #
# Products section                                                             #
#                                                                              #
################################################################################

## File structures give parameters for images (table 7 of LRIT TX SPEC)
_LRIT_PRODUCTS = (('full disk',             8, 1396, 698),
                  ('northern hemisphere',   8, 1396, 740),
                  ('southern hemisphere',   8, 1396, 740),
                  ('north west',            8, 1396, 740),
                  ('north east',            8, 1396, 698),
                  ('south west',            8, 1396, 690),
                  ('south east',            8, 1396, 740),
                  # environmental products
                  ('nws',                   8, 1200, 900),
                  ('gws',                   8, 1200, 804),
                  ('meteosat',              8, 1200, 804),
                 )
# NOTE: all _LRIT_PRODUCTS have 8 bits/pxl, whew no bit manipulation necessary
_ProductTuple = namedtuple('_ProductTuple', 'name, bits_per_pxl, ncols, nlines')
_PRODUCT_SPECS = (_ProductTuple(*tups) for tups in _LRIT_PRODUCTS)

################################################################################
#                                                                              #
# Features section                                                             #
#                                                                              #
################################################################################

import io
# class which encapsulates an Lrit file
class LritFile(io.FileIO):
    """Interface wrapper to LRIT file-like objects"""
    def __init__(self, name):
        super(LritFile, self).__init__(name, 'rb')   # read-only file
        self.primary_header = None

    have_primary_header = property(lambda self: self.primary_header is not None)

    # overload to read headers / data
    def next(self):
        """Returns the next LRIT file part, an instance of LritHeader or byte string"""
        # primary header must be first
        if not self.have_primary_header:
            self.primary_header = lrh.PrimaryHeader(self.read(lrh.PrimaryHeader.size))
            return self.primary_header

        # still reading headers
        if self.tell() < self.primary_header.tot_hdr_len:
            # first peek at the header type/len
            hpeek = lrh.LritHeader.parse_header(self.peek(lrh.MIN_HEADER_LEN))
            # now return the whole header based on the peeked hdr_type or unknown
            if self.tell() + hpeek.hdr_len > self.primary_header.tot_hdr_len:
                warnings.warn('Header appears to exceed total header length')
            buf = self.read(hpeek.hdr_len)
            try:
                return lrh.LritHeader.parse_header(buf, hpeek.hdr_type)
            except KeyError:
                return lrh.LritHeader.parse_header(buf, lrh.UNKNOWN_HTYPE)

        # return the data as raw byes, if subsequent read does not return 0,
        # there's unexpected extra data in this file
        data_bytes = self.primary_header.data_len // 8
        if data_bytes * 8 != self.primary_header.data_len:
            warnings.warn('Data field length is not an integer number of bytes')
        data = self.read(data_bytes)
        if len(data) > 0: return data
        raise StopIteration

    def peek(self, n):
        """Return n bytes read from file w/o advancing file pointer"""
        buf = self.read(n)
        self.seek(-len(buf), io.SEEK_CUR)
        return buf

# override builtin for lritfiles
def open(name): return LritFile(name)


# TODO: don't decompress in receiver, do that in these function
# TODO: Create a series of classes to translate headers to useful structures,
# e.g. the ImageDataHeader specifies descriptions of the image data and mappings
# from the image values to measured units, the ImageNavHeader describes the map
# projection of the image
# TODO: TextTuple can contain a "Time of frame start = YYYY/DDD/HH:MM:SS" string
# where YYYY is the year, DDD is the day of the year (from last day of prior year)
# and HH:MM:SS is the time.  This can be compared to any Timestamp value for a
# validation of some sort?


#### convenience functions ####
try:
    from PIL import Image
except ImportError:
    import Image  # TODO: Python Image Library dev has stagnated?  Switch to Pillow?

# tuple to contain file parts
LritFileParts = namedtuple('LritFileParts', 'headers, data')

def readfileparts(fn):
    """Returns some object (Image, string, etc.) given the LRIT filename"""
    headers = []
    with open(fn) as f:
        for part in f:
            if isinstance(part, lrh.LritHeader): headers.append(part)
    # last part is always data (or only headers?)
    if isinstance(part, lrh.LritHeader): return None
    return LritFileParts(headers, part)

def readfile(fn):
    """Returns Image object given the LRIT filename"""
    width, height, fmt = None, None, 'raw'
    with open(fn) as f:
        for part in f:
            if isinstance(part, lrh.ImageStructHeader):
                width, height, bpp = part.ncols, part.nlines, part.bits_per_pxl
            if isinstance(part, lrh.NOAALRITHeader):
                if part.iscompressed == 5:
                    fmt = 'gif'

    # last part is always the data
    if None in (width, height):
        warnings.warn('No image found in %s ' % fn)
        return None
    if fmt == 'raw' and width * height != len(part):
        warnings.warn('Invalid image dimensions %d x %d != %d' % (width, height, len(part)))
        return None
    # extended call parameters recommended by warning from Image
    args = tuple()#'L', 0, 1)
    return Image.frombuffer('L', (width, height), part, fmt, *args)   # should this be L (greyscale)  or RGB?

import tempfile
def stitchimage(files, fnout=None):
    """Returns file name for file stitched from files"""
    if fnout is not None:
        # rebuild fnout with jpg extension if none given
        fnout, ext = osp.splitext(fnout)
        fnout += ext or '.jpg'
        fn = btin.open(fnout, 'wb')
    else:
        fn = tempfile.NamedTemporaryFile(suffix='.jpg', delete=False)

    iparts = (readfileparts(f) for f in files)
    imgs = []
    rows = 0
    for part in iparts:
        img = [None, None]  # upper-left corner coords, image
        for h in part.headers:
            # save image
            if isinstance(h, lrh.ImageStructHeader):
                try:
                    img[0] = Image.frombuffer('L', (h.ncols, h.nlines), part.data, 'raw', 'L', 0, 1)
                except:
                    img[0] = Image.frombuffer('L', (h.ncols, len(part.data)/h.ncols), part.data, 'raw', 'L', 0, 1)

            # save offset
            if isinstance(h, lrh.SegmentIdHeader):
                #cols, rows = h.max_col, h.max_row
                img[1] = (h.col0, h.line0)
        imgs.append(img)

    if len(imgs) == 1:
        imgs[0][1] = (0,0)
    # find max col/row as maximum sum of columns and start column
    cols = max(x[1][0] + x[0].size[0] for x in imgs)
    rows = max(x[1][1] + x[0].size[1] for x in imgs)
    im = Image.new('L', (cols, rows), "black")
    for img in imgs: im.paste(*img)  # args are (Image object, offset)

    im.save(fn, 'JPEG')  # accepts file-like or string filename
    return fn.name   # file-like

from glob import glob, iglob
import re

def openemwin(FILE, iszip, output):
    suff = ".zip" if iszip else ".txt"
    fn = tempfile.NamedTemporaryFile(mode='w+b',suffix=suff, delete=False)
    part = readfileparts(FILE)
    fn.write(part.data)
    return fn.name

def imagefileglob(fn):
    """Return list of files belonging to the same image group as fn"""
    # need to use the SegmentIdHeader.id field
    with open(fn) as f:
        for part in f:
            if isinstance(part, lrh.SegmentIdHeader):
                segid = part.id
                break
        else:
            return [fn,]     # Not all images are segmented
    # file glob does not uniquely identify associated files, just reduces the search

    # old-school LRIT convention
    match = re.match(r'(.+seg)\d{3}', fn)
    globlist = []
    if match:
        globlist.extend(iglob(match.expand(r'\1*')))

    # HRIT undoc'd BASENAME_NSEG numbered segments
    match = re.match(r'(.+_)\d{3}$', osp.splitext(fn)[0])
    if match:
        globlist.extend(iglob(match.expand(r'\1*')))

    # GRB documented file segments
    #match = re.match(r'')
    match = re.match(r'(.+_)s\d{14}_e\d{14}_c\d{14}', osp.splitext(fn)[0])
    if match:
        globlist.extend(iglob(match.expand(r'\1*')))

    flist = []
    for gf in globlist:
        with open(gf) as f:
            for part in f:
                if isinstance(part, lrh.SegmentIdHeader) and part.id == segid:
                    flist.append(gf)
                    break
    return flist or [fn,]

################################################################################
#                                                                              #
# Tests and main section                                                       #
#                                                                              #
################################################################################

import sys
import argparse as argp

def get_argparser():
    pass

def viewfiles(*args):
    pass

import pprint as pp
def testmain(*args):
    im = readfile(args[0])
    if im is not None:
        im.save('imagefile.jpg', 'JPEG')
        print 'Saved image to imagefile.jpg'
    else:
        print 'No image'

if __name__ == '__main__':
#    sys.exit(viewfiles(*sys.argv[1:]))
    sys.exit(testmain(*sys.argv[1:]))
